# auto-generated file
__all__ = ['lib', 'ffi']

import os
from sourmash._lowlevel__ffi import ffi

lib = ffi.dlopen(os.path.join(os.path.dirname(__file__), '_lowlevel__lib.cp37-win_amd64.pyd'), 0)
del os
